
__version__ = "0.8.1"

__title__ = "jason-server"
__description__ = "JSON as REST API"
__url__ = "https://github.com/nekomamoushi/jason-server"

__author__ = "Jerome Pradier"
__author_email__ = "germione.grangere@gmail.com"

__license__ = "MIT"




from tinydb import TinyDB, Query
from tinydb.storages import MemoryStorage

from jason_server.utils import open_database

# --------------------------------------------------------------------------- #


def tinydb_make_table(db, name):
    return db.table(name)


def tinydb_populate_table(table, data):
    for elt in data:
        table.insert(elt)

# --------------------------------------------------------------------------- #


class Database(object):

    def __init__(self, path):
        self._path = path
        self._db = TinyDB(storage=MemoryStorage)
        self._json = open_database(path)
        self._endpoints = self._generate_tables()

    def _generate_tables(self):
        resources = [name for name in self._json]
        for name in resources:
            table = tinydb_make_table(self._db, name)
            tinydb_populate_table(table, self._json[name])
        return resources

    @property
    def json(self):
        return self._json

    @property
    def endpoints(self):
        return self._endpoints

    def resources(self):
        return [table for table in self._db.tables() if table != "_default"]

    def resource(self, name):
        return self._db.table(name)




from click.testing import CliRunner

from jason_server.cli import cli

NON_EXISTING_FILE = "tests/data/non_existing_file"
BAD_FORMATTED_FILE = "tests/data/bad_formatted_database.json"
SAMPLE_DATABASE = "tests/data/sample_database.json"


def describe_cli():

    def with_no_arguments():
        runner = CliRunner()
        result = runner.invoke(cli)
        assert 0 == result.exit_code


def describe_cli_watch():

    def with_non_existing_file():
        runner = CliRunner()
        result = runner.invoke(cli, ['watch', NON_EXISTING_FILE])
        assert 1 == result.exit_code
        assert NON_EXISTING_FILE in result.output

    def with_bad_formated_json():
        runner = CliRunner()
        result = runner.invoke(cli, ['watch', BAD_FORMATTED_FILE])
        assert 2 == result.exit_code
        assert BAD_FORMATTED_FILE in result.output




import webbrowser

import click

from jason_server.derulo import run

CONTEXT_SETTINGS = dict(help_option_names=['--help'])


@click.group(context_settings=CONTEXT_SETTINGS)
@click.option(
    "-h", "--host", default='localhost', help="Host adress"
)
@click.option(
    "-p", "--port", default=8080, help="Port"
)
@click.option(
    "-q", "--quiet", is_flag=True
)
@click.version_option(
    version='0.8.1'
)
@click.pass_context
def cli(ctx, host, port, quiet):
    ctx.obj = {}
    ctx.obj["host"] = host
    ctx.obj["port"] = port
    ctx.obj["quiet"] = quiet


@cli.command(help="Run your database as REST Api")
@click.argument('database')
@click.option(
    "-o", "--open", is_flag=True
)
@click.pass_context
def watch(ctx, database, open):
    if open:
        url = "http://{}:{}".format(ctx.obj["host"], ctx.obj["port"])
        webbrowser.open(url)
    run(ctx.obj, database=database)




from jason_server.utils import chunk_list, str_to_int


def test_chunk_list():

    d = [{'a': 1}, {'b': 2}, {'c': 3}, {'d': 4}, {'e': 5}, {'f': 6}, {'g': 7}]
    chunk_d = chunk_list(d)
    chunk_l = list(chunk_d)
    assert len(d) == len(chunk_l)

    chunk_d = chunk_list(d, chunk_size=2)
    chunk_l = list(chunk_d)
    assert 4 == len(chunk_l)

    chunk_d = chunk_list(d, chunk_size=3)
    chunk_l = list(chunk_d)
    assert 3 == len(chunk_l)
    assert 1 == len(chunk_l[2])


def test_str_to_int():
    string_no_digit = "jesuisunefougere"
    string_all_digit = "1234"
    string_some_digit = "ehy45"

    assert string_no_digit == str_to_int(string_no_digit)
    assert 1234 == str_to_int(string_all_digit)
    assert string_some_digit == str_to_int(string_some_digit)



import pytest

import requests
import time

URL_HOME = "http://127.0.0.1:8100"
URL_PERSONS = "http://127.0.0.1:8100/persons"

URL_PERSONS_FILTER_AGE_30 = "http://127.0.0.1:8100/persons?age=30"
URL_PERSONS_FILTER_AGE_30_MALE = "http://127.0.0.1:8100/persons?age=30&gender=male"
URL_PERSONS_FILTER_AGE_30_FEMALE = "http://127.0.0.1:8100/persons?age=30&gender=female"

URL_PERSONS_PAGE_1 = "http://127.0.0.1:8100/persons?_page=1"
URL_PERSONS_PAGE_2 = "http://127.0.0.1:8100/persons?_page=2"

URL_PERSONS_PAGE_1_LIMIT_3 = "http://127.0.0.1:8100/persons?_page=1&_limit=3"
URL_PERSONS_PAGE_2_LIMIT_3 = "http://127.0.0.1:8100/persons?_page=2&_limit=3"
URL_PERSONS_PAGE_3_LIMIT_3 = "http://127.0.0.1:8100/persons?_page=3&_limit=3"
URL_PERSONS_PAGE_4_LIMIT_3 = "http://127.0.0.1:8100/persons?_page=4&_limit=3"
URL_PERSONS_PAGE_5_LIMIT_3 = "http://127.0.0.1:8100/persons?_page=5&_limit=3"

URL_PERSONS_SORT_BY_AGE = "http://127.0.0.1:8100/persons?_sort=age"
URL_PERSONS_SORT_BY_NAME = "http://127.0.0.1:8100/persons?_sort=name"

URL_PERSONS_SORT_BY_AGE_PAGE_1 = "http://127.0.0.1:8100/persons?_sort=age&_page=1"
URL_PERSONS_SORT_BY_NAME_PAGE_2 = "http://127.0.0.1:8100/persons?_sort=name&_page=2"
URL_PERSONS_SORT_DESC_BY_AGE_PAGE_1 = "http://127.0.0.1:8100/persons?_sort=age&_order=desc&_page=1"
URL_PERSONS_SORT_BY_NAME_PAGE_2_LIMIT_4 = "http://127.0.0.1:8100/persons?_sort=name&_page=2&_limit=4"


def describe_endpoint():

    @pytest.fixture()
    def s():
        # Wait for jason_server
        time.sleep(1)

        s = requests.Session()
        yield s

    def get(s):

        r = s.get(URL_HOME)
        assert 200 == r.status_code

        r = s.get(URL_PERSONS)
        assert 200 == r.status_code
        assert 'application/json' == r.headers['content-type']
        data = r.json()
        assert 13 == len(data['data'])

    def get_with_filter(s):

        r = s.get(URL_PERSONS_FILTER_AGE_30)
        assert 200 == r.status_code
        data = r.json()
        assert 3 == len(data['data'])

        r = s.get(URL_PERSONS_FILTER_AGE_30_MALE)
        assert 200 == r.status_code
        data = r.json()
        assert 2 == len(data['data'])

        r = s.get(URL_PERSONS_FILTER_AGE_30_FEMALE)
        assert 200 == r.status_code
        data = r.json()
        assert 1 == len(data['data'])

    def get_with_paginate(s):

        r = s.get(URL_PERSONS_PAGE_1)
        assert 200 == r.status_code
        data = r.json()
        assert 10 == len(data['data'])

        r = s.get(URL_PERSONS_PAGE_2)
        assert 200 == r.status_code
        data = r.json()
        assert 3 == len(data['data'])

    def get_with_paginate_and_limit(s):

        r = s.get(URL_PERSONS_PAGE_1_LIMIT_3)
        assert 200 == r.status_code
        data = r.json()
        assert 3 == len(data['data'])

        r = s.get(URL_PERSONS_PAGE_2_LIMIT_3)
        assert 200 == r.status_code
        data = r.json()
        assert 3 == len(data['data'])

        r = s.get(URL_PERSONS_PAGE_3_LIMIT_3)
        assert 200 == r.status_code
        data = r.json()
        assert 3 == len(data['data'])

        r = s.get(URL_PERSONS_PAGE_4_LIMIT_3)
        assert 200 == r.status_code
        data = r.json()
        assert 3 == len(data['data'])

        r = s.get(URL_PERSONS_PAGE_5_LIMIT_3)
        assert 200 == r.status_code
        data = r.json()
        assert 1 == len(data['data'])

    def get_with_sorting(s):

        r = s.get(URL_PERSONS_SORT_BY_AGE)
        assert 200 == r.status_code
        data = r.json()
        assert 21 == data['data'][0]['age']
        assert "Bell Hinton" == data['data'][0]['name']

        r = s.get(URL_PERSONS_SORT_BY_NAME)
        assert 200 == r.status_code
        data = r.json()
        assert 30 == data['data'][0]['age']
        assert "Anastasia Coffey" == data['data'][0]['name']

    def get_with_sorting_and_pagination(s):

        r = s.get(URL_PERSONS_SORT_BY_AGE_PAGE_1)
        assert 200 == r.status_code
        data = r.json()
        assert 23 == data['data'][1]['age']

        r = s.get(URL_PERSONS_SORT_BY_NAME_PAGE_2)
        assert 200 == r.status_code
        data = r.json()
        assert 36 == data['data'][1]['age']

        r = s.get(URL_PERSONS_SORT_DESC_BY_AGE_PAGE_1)
        assert 200 == r.status_code
        data = r.json()
        assert 37 == data['data'][1]['age']

        r = s.get(URL_PERSONS_SORT_BY_NAME_PAGE_2_LIMIT_4)
        assert 200 == r.status_code
        data = r.json()
        assert 23 == data['data'][0]['age']



import os
from functools import reduce

import bottle
from bottle import Bottle, template, request, response
from tinydb import where

from jason_server.database import Database
from jason_server.utils import chunk_list, str_to_int

# --------------------------------------------------------------------------- #

# Define base directory
BASE_DIR = os.path.dirname(os.path.abspath(__file__))

# App Config
bottle.TEMPLATE_PATH.insert(0, os.path.join(BASE_DIR, 'views'))

app = Bottle()
db = None

# --------------------------------------------------------------------------- #


def build_link_header(request, page, total):

    REL_FIRST = '<{url}?_page={first}>; rel="first"'
    REL_PREV = '<{url}?_page={prev}>; rel="prev"'
    REL_NEXT = '<{url}?_page={next}>; rel="next"'
    REL_LAST = '<{url}?_page={last}>; rel="last"'

    scheme, netloc, path = request.urlparts[0:3]
    url = "{scheme}://{netloc}{path}".format(
        scheme=scheme,
        netloc=netloc,
        path=path
    )
    links = None

    if total == 1:
        return ""

    if page == 1:
        links = [
            REL_FIRST.format(url=url, first=1),
            REL_NEXT.format(url=url, next=2),
            REL_LAST.format(url=url, last=total)
        ]
    elif page == total:
        links = [
            REL_FIRST.format(url=url, first=1),
            REL_PREV.format(url=url, prev=total-1),
            REL_LAST.format(url=url, last=total)
        ]
        return ",".join(links)
    else:
        links = [
            REL_FIRST.format(url=url, first=1),
            REL_PREV.format(url=url, prev=page-1),
            REL_NEXT.format(url=url, next=page+1),
            REL_LAST.format(url=url, last=total)
        ]

    return ",".join(links)


def query_filter(table, arguments):
    user_arguments = [
        (k, v)
        for k, v in arguments.items()
        if not k.startswith('_')
    ]
    if not user_arguments:
        return table.all()

    query = [where(k) == str_to_int(v) for k, v in user_arguments]

    resources = table.search(reduce(lambda a, b: a & b, query))

    return resources


def query_sort(resources, arguments):
    if '_sort' not in arguments:
        return resources

    sort = arguments['_sort']
    order = 'asc' if '_order' not in arguments else arguments['_order']
    if order == 'asc':
        return sorted(resources, key=lambda i: i[sort])

    return sorted(resources, key=lambda i: i[sort], reverse=True)


def query_paginate(resources, arguments):
    if '_page' not in arguments:
        return resources

    page = int(arguments['_page'])
    limit = 10 if '_limit' not in arguments else int(arguments['_limit'])
    chunk_data = list(chunk_list(resources, limit))
    results = chunk_data[page-1]
    link_header = build_link_header(request, page, len(chunk_data))
    response.set_header("Link", link_header)
    return results


def retrieve_resources(endpoint, arguments=None):

    table = db.resource(endpoint)

    if not arguments:
        return dict(data=table.all())

    results = query_filter(table, request.query)
    results = query_sort(results, request.query)
    results = query_paginate(results, request.query)

    return dict(data=results)


def retrieve_resource(endpoint, index):
    table = db.resource(endpoint)
    elements = table.all()
    index = int(index) - 1

    if index >= len(elements):
        return {}

    return elements[index]


# --------------------------------------------------------------------------- #

@app.hook('after_request')
def set_default_headers():
    response.set_header("X-Powerded-By", "Bottle")


@app.route('/')
def bottle_world():

    def build_url(host, port):
        return "http://{}:{}".format(host, port)

    host = app.config.get('host')
    port = app.config.get('port')
    resources = {
        "base_url": build_url(host, port),
        "tables": db.endpoints
    }
    return template('index.html', resources)


@app.route('/db')
def db():
    return db.json


@app.route('/<endpoint>', method='GET')
def get_resources(endpoint):
    results = retrieve_resources(endpoint, request.query)
    return results


@app.route('/<endpoint>/<index>', method='GET')
def get_resource(endpoint, index):
    resource = retrieve_resource(endpoint, index)
    return resource

# --------------------------------------------------------------------------- #


def print_message(database, tables, host, port):
    BOLD_BLUE = '\033[1;34m'
    RESET = '\033[0m'

    print("\n    {}Database{}".format(BOLD_BLUE, RESET))
    print("    {}".format(database))

    print("\n    {}Resources{}".format(BOLD_BLUE, RESET))
    base_url = "http://{}:{}".format(host, port)
    for table in tables:
        print("    {}/{}".format(base_url, table))

    print("\n    {}Home{}".format(BOLD_BLUE, RESET))
    print("    {}/".format(base_url))


def run(options, database):
    host = options['host']
    port = options['port']
    quiet = options['quiet']

    app.config.setdefault('host', host)
    app.config.setdefault('port', port)
    global db
    db = Database(database)
    if not quiet:
        print_message(database, db.endpoints, host, port)
    app.run(host=host, port=port, quiet=True)




import os
import io

from setuptools import find_packages, setup

# Needed for the following functions
HERE = os.path.abspath(os.path.dirname(__file__))


# Load the package's __init__.py module as a dictionary.
ABOUT = {}
with io.open(os.path.join(HERE, 'src/jason_server', '__init__.py')) as f:
    exec(f.read(), ABOUT)


# Get the long description from the README file
def get_long_description():
    with io.open(os.path.join(HERE, 'README.md'), encoding='utf-8') as f:
        long_description = f.read()
    return long_description


PACKAGE_NAME = ABOUT['__title__']
VERSION = ABOUT['__version__']
DESCRIPTION = ABOUT['__description__']
URL = ABOUT['__url__']
AUTHOR = ABOUT['__author__']
AUTHOR_EMAIL = ABOUT['__author_email__']
LICENSE = ABOUT['__license__']
PACKAGES = find_packages(where='src', exclude=['contrib', 'docs', 'tests'])
PACKAGE_DIR = {"": "src"}
INSTALL_REQUIRES = [
    "click",
    "bottle",
    "tinydb"
]
EXTRA_REQUIRES = []

###############################################################################

setup(
    # Name of the package
    name=PACKAGE_NAME,

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=VERSION,

    description=DESCRIPTION,
    long_description=get_long_description(),
    long_description_content_type='text/markdown',

    # The project's main homepage.
    url=URL,

    # Author details
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,

    # Choose your license
    # Don't forget to the license file
    license=LICENSE,

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
    ],

    # What does your project relate to?
    keywords='packaging setuptools development',

    # Adding Non-Code Files :
    # In order for these files to be copied at install time to the package’s folder inside site-packages,
    # you’ll need to supply :
    include_package_data=True,

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=PACKAGES,
    package_dir=PACKAGE_DIR,

    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=INSTALL_REQUIRES,

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    # See https://setuptools.readthedocs.io/en/latest/setuptools.html#automatic-script-creation
    entry_points=dict(
        console_scripts=[
            'jason-server = jason_server.cli : cli'
        ]
    )
)




from jason_server.database import Database

NON_EXISTING_FILE = "tests/data/non_existing_file"
BAD_FORMATTED_FILE = "tests/data/bad_formatted_database.json"
JSON_DATABASE = "tests/data/sample_database.json"


def describe_database():

    def with_existing_database():
        db = Database(path=JSON_DATABASE)
        assert 2 == len(db.endpoints)
        assert 'articles' in db.endpoints
        assert 'authors' in db.endpoints
        assert 2 == len(db.resources())
        assert 2 == len(db.resource('articles'))
        assert 3 == len(db.resource('authors'))




from itertools import islice
import json

try:
    FileNotFoundError
except:
    FileNotFoundError = IOError

try:
    from json.decoder import JSONDecodeError
except ImportError:
    JSONDecodeError = ValueError


def open_database(path):
    try:
        with open(path, "r") as f:
            return json.load(f)
    except FileNotFoundError:
        print("< {0} > does not exists.".format(path))
        exit(1)
    except JSONDecodeError:
        print("< {0} > is not well formated.".format(path))
        exit(2)


def chunk_list(data, chunk_size=1):
    it = iter(data)
    return iter(lambda: tuple(islice(it, chunk_size)), ())


def str_to_int(value):
    try:
        return int(value)
    except ValueError:
        return value
